package at.tugraz.iaik.p2p.commons

import iaik.asn1.structures.AlgorithmID
import iaik.security.ec.common.ECParameterSpec
import iaik.security.ec.common.ECStandardizedParameterFactory
import iaik.security.ec.common.PointEncoders
import iaik.security.ec.errorhandling.DecodingException
import iaik.security.ec.errorhandling.InvalidCurveParameterSpecException
import iaik.security.ec.provider.ECCelerate
import iaik.security.provider.IAIK
//import iaik.security.ssl.ECCelerateProvider
//import iaik.security.ssl.SecurityProvider
import iaik.x509.X509Certificate
import iaik.x509.X509ExtensionException
import iaik.x509.extensions.BasicConstraints
import iaik.x509.extensions.CRLDistributionPoints
import iaik.x509.extensions.ExtendedKeyUsage
import iaik.x509.extensions.KeyUsage
import iaik.x509.extensions.netscape.NetscapeCertType
import java.io.*
import java.math.BigInteger
import java.security.*
import java.security.interfaces.ECPrivateKey
import java.security.interfaces.ECPublicKey
import java.util.*
import javax.crypto.Cipher
import kotlin.experimental.and
import kotlin.experimental.or

/**
 * Created by Bernd Prünster <bernd.pruenster@iaik.tugraz.at>
 */
object CommonCrypto {
    val TLS_CIPHERS: Array<String> get() = getCipherSuites("EC") + getCipherSuites("RSA")
    //private static final Object LOCK = new Object();
    val RND: SecureRandom = SecureRandom.getInstance("SHA512PRNG-SP80090", IAIK.getInstance());

    init {
//        Security.removeProvider("SUN")
//        Security.removeProvider("SunEC")
//        Security.removeProvider("SunRsaSign")
//        Security.removeProvider("SunJCE")
//        Security.removeProvider("SunJSSE")
//        Security.removeProvider("SunJGSS")
//        Security.removeProvider("SunPCSC")
        IAIK.addAsProvider(true)
        ECCelerate.addAsProvider(true);
        //  SecurityProvider.setSecurityProvider(ECCelerateProvider())
        println(Security.getProviders())
    }


    @Throws(CryptoException::class)
    @JvmStatic
    fun genKP(algorithm: String, len: Int): KeyPair {
        val ec: KeyPairGenerator
        try {
            ec = KeyPairGenerator.getInstance(algorithm,
                                              if (algorithm == "EC") ECCelerate.getInstance() else IAIK.getInstance())
            if (algorithm == "EC") ec.initialize(ECStandardizedParameterFactory.getParametersByBitLength(len), RND)
            else ec.initialize(len)
            return ec.generateKeyPair()
        } catch (e: InvalidAlgorithmParameterException) {
            throw CryptoException(e)
        } catch (e: NoSuchAlgorithmException) {
            throw CryptoException(e)
        }

    }

    @Throws(CryptoException::class)
    @JvmStatic
    fun genKP(): ECKeyPair {
        val ec: KeyPairGenerator
        try {
            ec = KeyPairGenerator.getInstance("EC", ECCelerate.getInstance())
            ec.initialize(ECStandardizedParameterFactory.getParametersByBitLength(256), RND)
            return ECKeyPair(ec.generateKeyPair())
        } catch (e: InvalidAlgorithmParameterException) {
            throw CryptoException(e)
        } catch (e: NoSuchAlgorithmException) {
            throw CryptoException(e)
        }

    }

    @Throws(CryptoException::class)
    @JvmStatic
    fun genCert(pub: PublicKey,
                subject: Principal,
                issuerCN: Principal = subject,
                signingKey: PrivateKey,
                ca: Boolean,
                from: Date,
                to: Date,
                crlDistributionPoints: Array<String>? = null): java.security.cert.X509Certificate {

        val cert = X509Certificate()
        cert.serialNumber = BigInteger.valueOf(RND.nextLong())
        cert.subjectDN = subject
        try {
            cert.publicKey = pub

        } catch (e: InvalidKeyException) {
            throw CryptoException(e)
        }

        cert.issuerDN = issuerCN
        cert.setValidNotBefore(from)
        cert.setValidNotAfter(to)
        // CRL distribution point
        crlDistributionPoints?.let {
            if (!crlDistributionPoints.isEmpty()) {
                val point = iaik.asn1.structures.DistributionPoint(crlDistributionPoints)
                point.reasonFlags = iaik.asn1.structures.DistributionPoint.keyCompromise
                cert.addExtension(CRLDistributionPoints(point))
            }
        }

        try {
            if (ca) {
                val basicConstraints = BasicConstraints(true, 1)
                basicConstraints.isCritical = true
                cert.addExtension(basicConstraints)
                cert.addExtension(NetscapeCertType(NetscapeCertType.SSL_CA or NetscapeCertType.S_MIME_CA))

                val keyUsage = KeyUsage(KeyUsage.digitalSignature or KeyUsage.cRLSign or KeyUsage.keyCertSign)
                keyUsage.isCritical = true
                cert.addExtension(keyUsage)
            } else {
                cert.addExtension(KeyUsage(KeyUsage.digitalSignature or KeyUsage.keyEncipherment))
                cert.addExtension(ExtendedKeyUsage(arrayOf(ExtendedKeyUsage.serverAuth, ExtendedKeyUsage.clientAuth)))
            }
        } catch (e: X509ExtensionException) {
            throw CryptoException(e)
        }

        try {
            cert.sign(AlgorithmID.getAlgorithmID(signingKey.getSignatureAlgorithm()), signingKey)
        } catch (e: Exception) {
            throw CryptoException(e)
        }

        return cert
    }


    @Throws(CryptoException::class)
    @JvmStatic
    fun encode(pubKey: ECPublicKey): ByteArray {
        val params: ECParameterSpec
        try {
            params = ECParameterSpec.getParameterSpec(pubKey.params)
        } catch (e: InvalidCurveParameterSpecException) {
            throw CryptoException(e)
        }

        val compressedPoint = params.curve.encodePoint(pubKey.w, PointEncoders.COMPRESSED)
        compressedPoint[0] = ((compressedPoint[0] and 1.toByte()) or 2.toByte())
        return compressedPoint
    }

    @Throws(CryptoException::class)
    @JvmStatic
    fun decode(encoded: ByteArray): ECPublicKey {
        val params: ECParameterSpec
        try {
            params = ECParameterSpec.getParameterSpec(specFromPubKey(encoded))
        } catch (e: InvalidCurveParameterSpecException) {
            throw CryptoException(e)
        }

        encoded[0] = (encoded[0] or 2.toByte())

        try {
            val ecPoint = params.curve.decodePoint(encoded)
            return iaik.security.ec.common.ECPublicKey(params, ecPoint)
        } catch (e: DecodingException) {
            throw CryptoException(e)
        }

    }

    @Throws(CryptoException::class) private fun specFromPubKey(encodedKey: ByteArray): java.security.spec.ECParameterSpec {
        val typ = (encodedKey[0] and 254.toByte())
        when (typ) {
            2.toByte() -> return ECStandardizedParameterFactory.getParametersByBitLength(256)
            else       -> throw CryptoException("Unknown Key Type: $typ")
        }
    }


    @Throws(CryptoException::class)
    @JvmStatic
    fun encrypt(key: ECPublicKey, input: ByteArray): ByteArray {
        val cipher: Cipher
        try {
            cipher = Cipher.getInstance("ECIES")
            cipher.init(Cipher.ENCRYPT_MODE, key)
            return cipher.doFinal(input)
        } catch (e: Exception) {
            throw CryptoException(e)
        }

    }

    @Throws(CryptoException::class)
    @JvmStatic
    fun sign(key: ECPrivateKey, data: ByteArray): ByteArray {
        try {

            val bOut = ByteArrayOutputStream()
            val out = DataOutputStream(bOut)
            val signed: ByteArray
            // synchronized (LOCK) {
            val signature = Signature.getInstance("SHA3-256withECDSA")
            signature.initSign(key)
            signature.update(data)
            signed = signature.sign()
            // }
            out.writeInt(data.size)
            out.write(data)
            out.writeInt(signed.size)
            out.write(signed)
            out.close()
            return bOut.toByteArray()
        } catch (e: Exception) {
            throw CryptoException(e)
        }

    }

    @Throws(CryptoException::class)
    @JvmStatic
    fun validate(data: ByteArray, key: ECPublicKey): ByteArray {
        try {
            val bIn = ByteArrayInputStream(data)
            val dIn = DataInputStream(bIn)
            val dataLen = dIn.readInt()
            val payLoad = ByteArray(dataLen)
            dIn.readFully(payLoad)

            val sigLen = dIn.readInt()
            val sigBytes = ByteArray(sigLen)
            // synchronized (LOCK) {
            val signature = Signature.getInstance("SHA3-256withECDSA")
            signature.initVerify(key)
            signature.update(payLoad)
            dIn.readFully(sigBytes)
            if (!signature.verify(sigBytes)) {
                // System.err.print(sigBytes.length + ": " + Arrays.toString(sigBytes));
                throw SignatureException("Invalid signature")
            }
            //}
            return payLoad
        } catch (e: InvalidKeyException) {
            throw CryptoException(e)
        } catch (e: NoSuchAlgorithmException) {
            throw CryptoException(e)
        } catch (e: IOException) {
            throw CryptoException(e)
        } catch (e: SignatureException) {
            throw CryptoException(e)
        }

    }


    @Throws(CryptoException::class)
    @JvmStatic
    fun sha3(data: ByteArray): ByteArray {
        val sha3: MessageDigest = MessageDigest.getInstance("SHA3-256")
        sha3.update(data)
        return sha3.digest()
    }

    @JvmStatic
    fun init() {
        sha3(byteArrayOf(0))
    }
}

class ECKeyPair internal constructor(kp: KeyPair) {
    /**
     * Returns a reference to the public key component of this key pair.
     *
     * @return a reference to the public key.
     */
    val public: ECPublicKey = kp.public as ECPublicKey
    /**
     * Returns a reference to the private key component of this key pair.
     *
     * @return a reference to the private key.
     */
    val private: ECPrivateKey = kp.private as ECPrivateKey


    fun toKP(): KeyPair {
        return KeyPair(public, private)
    }

}


fun PrivateKey.getSignatureInstance(): Signature {
    return Signature.getInstance(getSignatureAlgorithm())
}

fun PrivateKey.getSignatureAlgorithm(): String {
    if (algorithm == "RSA") return "SHA3-256/RSA"
    if (algorithm == "EC") return "SHA3-256/ECDSA"
    throw CryptoException("Algorithm $algorithm is not supported!")
}

fun PrivateKey.getTLSCipherSuites(): Array<String> {
    return getCipherSuites(algorithm)
}

fun PublicKey.getTLSCipherSuites(): Array<String> {
    return getCipherSuites(algorithm)
}


private fun getCipherSuites(algorithm: String): Array<String> {
    val alg: String
    if (algorithm == "RSA") alg = "RSA"
    else if (algorithm == "EC") alg = "ECDSA"
    else throw CryptoException("Algorithm $algorithm not supported!")

    return arrayOf("TLS_ECDHE_${alg}_WITH_AES_256_CBC_SHA384",
                   "TLS_ECDHE_${alg}_WITH_AES_128_GCM_SHA256",
                   "TLS_ECDHE_${alg}_WITH_AES_128_CBC_SHA256")


}

data class KeyAndCertChain(val key: PrivateKey, val chain: Array<java.security.cert.X509Certificate>)

class CryptoException(msg: String? = null, e: Exception? = null) : Exception(msg, e) {
    constructor(e: Exception) : this(null, e)
}
