package at.tugraz.iaik.p2p.kademlia;

import at.tugraz.iaik.p2p.kademlia.dht.DHT;
import at.tugraz.iaik.p2p.kademlia.dht.KademliaDHT;
import at.tugraz.iaik.p2p.kademlia.exceptions.RoutingException;
import at.tugraz.iaik.p2p.kademlia.message.Message;
import at.tugraz.iaik.p2p.kademlia.message.MessageFactory;
import at.tugraz.iaik.p2p.kademlia.message.Receiver;
import at.tugraz.iaik.p2p.kademlia.net.Connector;
import at.tugraz.iaik.p2p.kademlia.net.UdpConnector;
import at.tugraz.iaik.p2p.kademlia.node.Identifier;
import at.tugraz.iaik.p2p.kademlia.node.Node;
import at.tugraz.iaik.p2p.kademlia.node.SHA3ID;
import at.tugraz.iaik.p2p.kademlia.operation.ConnectOperation;
import at.tugraz.iaik.p2p.kademlia.operation.LookupForDataOperation;
import at.tugraz.iaik.p2p.kademlia.operation.NodeLookupOperation;
import at.tugraz.iaik.p2p.kademlia.operation.Operation;
import at.tugraz.iaik.p2p.kademlia.routing.JKademliaRoutingTable;
import at.tugraz.iaik.p2p.kademlia.routing.KademliaRoutingTable;
import at.tugraz.iaik.p2p.kademlia.util.serializer.JsonDHTSerializer;
import at.tugraz.iaik.p2p.kademlia.util.serializer.JsonRoutingTableSerializer;
import at.tugraz.iaik.p2p.kademlia.util.serializer.JsonSerializer;

import java.io.*;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;

/**
 * The main Kademlia Node on the network, this node manages everything for this local system.
 *
 * @author Joshua Kissoon
 * @todo When we receive a store message - if we have a newer version of the content, re-send this newer version to
 * that node so as to update their version
 * @todo Handle IPv6 Addresses
 * @since 20140215
 */
public class JKademliaNode implements KademliaNode {

    /* Kademlia Attributes */
    private final String ownerId;

    /* Objects to be used */
    private final transient Node                 localNode;
    private final transient KadServer            server;
    private final transient KademliaDHT          dht;
    private final transient KademliaRoutingTable routingTable;
    private final transient KadConfiguration     config;

    /* Timer used to execute refresh operations */
    private transient Timer     refreshOperationTimer;
    private transient TimerTask refreshOperationTTask;

    /* Factories */
    private final transient MessageFactory messageFactory;

    /* Statistics */
    private final transient KadStatistician statistician;


    {
        this.statistician = new Statistician();
    }

    /**
     * Creates a Kademlia DistributedMap using the specified name as filename base.
     * If the id cannot be read from disk the specified defaultId is used.
     * The instance is bootstraped to an existing network by specifying the
     * address of a bootstrap node in the network.
     *
     * @param ownerId      The Name of this node used for storage
     * @param localNode    The Local Node for this Kad instance
     * @param dht          The DHT for this instance
     * @param routingTable
     * @param config
     * @throws IOException If an error occurred while reading id or local map
     *                     from disk <i>or</i> a network error occurred while
     *                     attempting to bootstrap to the network
     */
    public JKademliaNode(final String ownerId,
                         final Node localNode,
                         final KademliaDHT dht,
                         final KademliaRoutingTable routingTable,
                         final KadConfiguration config) throws IOException {
        this.ownerId = ownerId;
        this.localNode = localNode;
        this.dht = dht;
        this.config = config;
        this.routingTable = routingTable;
        this.messageFactory = new MessageFactory(this, this.dht, this.config);
        this.server = new KadServer(this.messageFactory, this.localNode, this.config, this.statistician);
//        this.startRefreshOperation();
    }

//    @Override
//    public final void startRefreshOperation() {
//        this.refreshOperationTimer = new Timer(true);
//        this.refreshOperationTTask = new TimerTask() {
//            @Override
//            public void run() {
//                try {
//                    /* Runs a DHT RefreshOperation  */
//                    JKademliaNode.this.refresh();
//                } catch (final IOException e) {
//                    System.err.println("KademliaNode: Refresh Operation Failed; Message: " + e.getMessage());
//                }
//            }
//        };
//        this.refreshOperationTimer.schedule(this.refreshOperationTTask, this.config.restoreInterval(), this.config
//                .restoreInterval());
//    }
//
//    @Override
//    public final void stopRefreshOperation() {
//        /* Close off the timer tasks */
//        this.refreshOperationTTask.cancel();
//        this.refreshOperationTimer.cancel();
//        this.refreshOperationTimer.purge();
//    }

    public JKademliaNode(final String ownerId,
                         final Node node,
                         final KademliaRoutingTable routingTable,
                         final KadConfiguration config) throws IOException {
        this(ownerId, node, new DHT(ownerId, config), routingTable, config);
    }

    public JKademliaNode(final String ownerId, final Node node, final KadConfiguration config) throws IOException {
        this(ownerId, node, new JKademliaRoutingTable(node, config), config);
    }

    public JKademliaNode(final String ownerId, final Identifier defaultId, final int port) throws IOException {
        this(ownerId, defaultId, (short) 0, port);
    }

    public JKademliaNode(final String ownerId,
                         final Identifier defaultId,
                         final short nonceRange,
                         final int port) throws IOException {
        this(ownerId, defaultId, nonceRange, new UdpConnector(port));
    }

    public JKademliaNode(final String ownerId,
                         final Identifier defaultId,
                         final short nonceRange,
                         final Connector connector) throws IOException {
        this(ownerId, new Node(defaultId, nonceRange, connector), new DefaultConfiguration());
    }

    /**
     * Load Stored state using default configuration
     *
     * @param ownerId The ID of the owner for the stored state
     * @return A Kademlia instance loaded from a stored state in a file
     * @throws java.io.FileNotFoundException
     * @throws java.lang.ClassNotFoundException
     */
    public static JKademliaNode loadFromFile(final String ownerId) throws
                                                                   FileNotFoundException,
                                                                   IOException,
                                                                   ClassNotFoundException {
        return JKademliaNode.loadFromFile(ownerId, new DefaultConfiguration());
    }

    /**
     * Load Stored state
     *
     * @param ownerId The ID of the owner for the stored state
     * @param iconfig Configuration information to work with
     * @return A Kademlia instance loaded from a stored state in a file
     * @throws java.io.FileNotFoundException
     * @throws java.lang.ClassNotFoundException
     */
    public static JKademliaNode loadFromFile(final String ownerId, final KadConfiguration iconfig) throws
                                                                                                   FileNotFoundException,
                                                                                                   IOException,
                                                                                                   ClassNotFoundException {
        DataInputStream din;

        /**
         * @section Read Basic Kad data
         */
        din =
                new DataInputStream(new FileInputStream(getStateStorageFolderName(ownerId, iconfig) +
                                                        File.separator +
                                                        "kad.kns"));
        final JKademliaNode ikad = new JsonSerializer<JKademliaNode>().read(din);

        /**
         * @section Read the routing table
         */
        din =
                new DataInputStream(new FileInputStream(getStateStorageFolderName(ownerId, iconfig) +
                                                        File.separator +
                                                        "routingtable.kns"));
        final KademliaRoutingTable irtbl = new JsonRoutingTableSerializer(iconfig).read(din);

        /**
         * @section Read the node state
         */
        din =
                new DataInputStream(new FileInputStream(getStateStorageFolderName(ownerId, iconfig) +
                                                        File.separator +
                                                        "node.kns"));
        final Node inode = new JsonSerializer<Node>().read(din);

        /**
         * @section Read the DHT
         */
        din =
                new DataInputStream(new FileInputStream(getStateStorageFolderName(ownerId, iconfig) +
                                                        File.separator +
                                                        "dht.kns"));
        final KademliaDHT idht = new JsonDHTSerializer().read(din);
        idht.setConfiguration(iconfig);

        return new JKademliaNode(ownerId, inode, idht, irtbl, iconfig);
    }

    @Override
    public Node getNode() {
        return this.localNode;
    }

    @Override
    public KadServer getServer() {
        return this.server;
    }

    @Override
    public KademliaDHT getDHT() {
        return this.dht;
    }

    @Override
    public KadConfiguration getCurrentConfiguration() {
        return this.config;
    }

    @Override
    public synchronized final void bootstrap(final Node n) throws IOException, RoutingException {
        final long      startTime = System.nanoTime();
        final Operation op        = new ConnectOperation(this.server, this, n, this.config);
        op.execute();
        final long endTime = System.nanoTime();
        this.statistician.setBootstrapTime(endTime - startTime);
    }

    @Override
    public synchronized int sendMessage(final Message msg, final Receiver recv) throws IOException {
        if (msg.getTarget().getConnector().getDestinationKnown()) {
            return server.sendMessage(msg, recv);
        }
        Identifier    id      = msg.getTarget().getNodeId();
        List<Message> closest = routingTable.findClosest(id, 1);
        if (closest.get(0).getOrigin().getNodeId().equals(id)) {
            msg.getTarget().setConnector(closest.get(0).getOrigin().getConnector());
            return server.sendMessage(msg, recv);
        }

        NodeLookupOperation nlop = new NodeLookupOperation(server, this, id, config);
        nlop.execute();
        if (nlop.getClosestNodes().get(0).getNodeId().equals(id)) {
            msg.getTarget().setConnector(closest.get(0).getOrigin().getConnector());
            return server.sendMessage(msg, recv);
        }

        throw new IOException("Could not find node " + id);
    }

    public synchronized List<Node> getNodesForData(SHA3ID id) throws IOException {
        LookupForDataOperation lookup = new LookupForDataOperation(server, this, id, config);
        lookup.execute();
        return lookup.getClosestNodes();
    }


//    @Override
//    public int put(final KadContent content) throws IOException {
//        return this.put(new JKademliaStorageEntry(content));
//    }
//
//    @Override
//    public int put(final JKademliaStorageEntry entry) throws IOException {
//        final StoreOperation sop = new StoreOperation(this.server, this, entry, this.dht, this.config);
//        sop.execute();
//
//        /* Return how many nodes the content was stored on */
//        return sop.numNodesStoredAt();
//    }
//
//    @Override
//    public void putLocally(final KadContent content) throws IOException {
//        this.dht.store(new JKademliaStorageEntry(content));
//    }

//    @Override
//    public JKademliaStorageEntry get(final GetParameter param) throws NoSuchElementException, IOException,
//            ContentNotFoundException {
//        if (this.dht.contains(param)) {
//            /* If the content exist in our own DHT, then return it. */
//            return this.dht.get(param);
//        }
//
//        /* Seems like it doesn't exist in our DHT, get it from other Nodes */
//        final long startTime = System.nanoTime();
//        final ContentLookupOperation clo = new ContentLookupOperation(this.server, this, param, this.config);
//        clo.execute();
//        final long endTime = System.nanoTime();
//        this.statistician.addContentLookup(endTime - startTime, clo.routeLength(), clo.isContentFound());
//        return clo.getContentFound();
//    }
//
//    @Override
//    public void refresh() throws IOException {
//        new KadRefreshOperation(this.server, this, this.dht, this.config).execute();
//    }

    @Override
    public String getOwnerId() {
        return this.ownerId;
    }

    @Override
    public Connector getConnector() {
        return this.localNode.getConnector();
    }

    @Override
    public void shutdown(final boolean saveState) throws IOException {
        /* Shut down the server */
        this.server.shutdown();

//        this.stopRefreshOperation();

        /* Save this Kademlia instance's state if required */
        if (saveState) {
            /* Save the system state */
            this.saveKadState();
        }
    }

    @Override
    public void saveKadState() throws IOException {
        DataOutputStream dout;

        /**
         * @section Store Basic Kad data
         */
        dout =
                new DataOutputStream(new FileOutputStream(getStateStorageFolderName(this.ownerId, this.config) +
                                                          File.separator +
                                                          "kad.kns"));
        new JsonSerializer<JKademliaNode>().write(this, dout);

        /**
         * @section Save the node state
         */
        dout =
                new DataOutputStream(new FileOutputStream(getStateStorageFolderName(this.ownerId, this.config) +
                                                          File.separator +
                                                          "node.kns"));
        new JsonSerializer<Node>().write(this.localNode, dout);

        /**
         * @section Save the routing table
         * We need to save the routing table separate from the node since the routing table will contain the node and
         * the node will contain the routing table
         * This will cause a serialization recursion, and in turn a Stack Overflow
         */
        dout =
                new DataOutputStream(new FileOutputStream(getStateStorageFolderName(this.ownerId, this.config) +
                                                          File.separator +
                                                          "routingtable.kns"));
        new JsonRoutingTableSerializer(this.config).write(this.getRoutingTable(), dout);

        /**
         * @section Save the DHT
         */
        dout =
                new DataOutputStream(new FileOutputStream(getStateStorageFolderName(this.ownerId, this.config) +
                                                          File.separator +
                                                          "dht.kns"));
        new JsonDHTSerializer().write(this.dht, dout);

    }

    /**
     * Get the name of the folder for which a content should be stored
     *
     * @return String The name of the folder to store node states
     */
    private static String getStateStorageFolderName(final String ownerId, final KadConfiguration iconfig) {
        /* Setup the nodes storage folder if it doesn't exist */
        final String path            = iconfig.getNodeDataFolder(ownerId) + File.separator + "nodeState";
        final File   nodeStateFolder = new File(path);
        if (!nodeStateFolder.isDirectory()) {
            nodeStateFolder.mkdir();
        }
        return nodeStateFolder.toString();
    }

    @Override
    public KademliaRoutingTable getRoutingTable() {
        return this.routingTable;
    }

    @Override
    public KadStatistician getStatistician() {
        return this.statistician;
    }

    /**
     * Creates a string containing all data about this Kademlia instance
     *
     * @return The string representation of this Kad instance
     */
    @Override
    public String toString() {
        final StringBuilder sb = new StringBuilder("\n\nPrinting Kad State for instance with owner: ");
        sb.append(this.ownerId);
        sb.append("\n\n");

        sb.append("\n");
        sb.append("Local Node");
        sb.append(this.localNode);
        sb.append("\n");

        sb.append("\n");
        sb.append("Routing Table: ");
        sb.append(this.getRoutingTable());
        sb.append("\n");

        sb.append("\n");
        sb.append("DHT: ");
        sb.append(this.dht);
        sb.append("\n");

        sb.append("\n\n\n");

        return sb.toString();
    }
}
