package at.tugraz.iaik.p2p.kademlia;

import at.tugraz.iaik.p2p.kademlia.exceptions.KadServerDownException;
import at.tugraz.iaik.p2p.kademlia.message.MessageFactory;
import at.tugraz.iaik.p2p.kademlia.net.MessageMapping;
import at.tugraz.iaik.p2p.kademlia.node.Node;
import at.tugraz.iaik.p2p.kademlia.message.Message;
import at.tugraz.iaik.p2p.kademlia.message.Receiver;

import java.io.IOException;
import java.security.SecureRandom;
import java.util.HashMap;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;

/**
 * The server that handles sending and receiving messages between nodes on the Kad Network
 *
 * @author Joshua Kissoon
 * @created 20140215
 */
public class KadServer {


    /* Basic Kad Objects */
    private final transient KadConfiguration config;

    /* Server Objects */
    private transient boolean isRunning;
    private final Map<Integer, Receiver> receivers;
    private final Timer timer;      // Schedule future tasks
    private final Map<Integer, TimerTask> tasks;    // Keep track of scheduled tasks

    private final Node localNode;

    /* Factories */
    private final MessageFactory messageFactory;

    private final KadStatistician statistician;
    public final SecureRandom random = new SecureRandom();


    {
        this.isRunning = true;
        this.tasks = new HashMap<>();
        this.receivers = new HashMap<>();
        this.timer = new Timer(true);
    }

    /**
     * Initialize our KadServer
     *
     * @param mFactory     Factory used to create messages
     * @param localNode    Local node on which this server runs on
     * @param config
     * @param statistician A statistician to manage the server statistics
     * @throws java.net.SocketException
     */
    public KadServer(final MessageFactory mFactory, final Node localNode, final KadConfiguration config, final
    KadStatistician statistician) throws IOException {
        this.config = config;
        this.localNode = localNode;
        this.messageFactory = mFactory;
        this.statistician = statistician;
        localNode.getConnector().open();

        /* Start listening for incoming requests in a new thread */
        this.startListener();
    }

    /**
     * Starts the listener to listen for incoming messages
     */
    private void startListener() {
        new Thread() {
            @Override
            public void run() {
                setName("KadServer " + KadServer.this.localNode.getNodeId());
                listen();
            }
        }.start();
    }

    /**
     * Sends a message
     *
     * @param msg  The message to send
     * @param recv The receiver to handle the response message
     * @return Integer The communication ID of this message
     * @throws IOException
     * @throws KadServerDownException
     */
    public synchronized int sendMessage(final Message msg, final Receiver recv) throws IOException,
            KadServerDownException {
        if (!this.isRunning) {
            throw new KadServerDownException(this.localNode + " - Kad Server is not running.");
        }

        /* Generate a random communication ID */

        final int comm = msg.getCommID();

        /* If we have a receiver */
        if (recv != null) {
            try {
                /* Setup the receiver to handle message response */
                this.receivers.put(comm, recv);
                final TimerTask task = new TimeoutTask(comm, recv);
                this.timer.schedule(task, this.config.responseTimeout());
                this.tasks.put(comm, task);
            } catch (final IllegalStateException ex) {
                /* The timer is already cancelled so we cannot do anything here really */
            }
        }

        /* Send the message */
        sendMessage(msg);

        return comm;
    }

    /**
     * Method called to reply to a message received
     *
     * @param msg The reply message
     * @throws java.io.IOException
     */
    public synchronized void reply(final Message msg) throws IOException {
        if (!this.isRunning) {
            throw new IllegalStateException("Kad Server is not running.");
        }
        sendMessage(msg);
    }

    /**
     * Internal sendMessage method called by the public sendMessage method after a communicationId is generated
     */
    public void sendMessage(final Message msg) throws IOException {
        this.localNode.getConnector().send(msg);
    }

    /**
     * Listen for incoming messages in a separate thread
     */
    private void listen() {
        try {
            while (this.isRunning) {
                try {

                    /* We've received a packet, now handle it */

                    final MessageMapping data = this.localNode.getConnector().receive();
                    if (!data.getMsg().getTarget().equals(localNode)) {
                        System.err.println("discarding message: " + data.getMsg());
                        continue;
                    }
                        /* Read in the conversation Id to know which handler to handle this response */
                    final int comm = data.getId();

                    final Message msg = data.getMsg();

                        /* Get a receiver for this message */
                    final Receiver receiver;
                    if (this.receivers.containsKey(comm)) {
                            /* If there is a reciever in the receivers to handle this */
                        synchronized (this) {
                            receiver = this.receivers.remove(comm);
                            final TimerTask task = (TimerTask) this.tasks.remove(comm);
                            if (task != null) {
                                task.cancel();
                            }
                        }
                    } else {
                            /* There is currently no receivers, try to get one */
                        receiver = this.messageFactory.createReceiver(msg.code(), this);
                    }

                        /* Invoke the receiver */
                    if (receiver != null) {
                        receiver.receive(msg, comm);
                    }

                } catch (final IOException e) {
                    e.printStackTrace();
                    //this.isRunning = false;
//                    System.err.println("Server ran into a problem in listener method. Message: " + e.getMessage());
                }
            }
        } finally {
            try {
                this.localNode.getConnector().close();
            } catch (final IOException e) {
                e.printStackTrace();
            }
            this.isRunning = false;
        }
    }

    /**
     * Remove a conversation receiver
     *
     * @param comm The id of this conversation
     */
    private synchronized void unregister(final int comm) {
        this.receivers.remove(comm);
        this.tasks.remove(comm);
    }

    /**
     * Stops listening and shuts down the server
     */
    public synchronized void shutdown() {
        this.isRunning = false;
        try {
            this.localNode.getConnector().close();
        } catch (final IOException e) {
            e.printStackTrace();
        }
        this.timer.cancel();
    }

    /**
     * Task that gets called by a separate thread if a timeout for a receiver occurs.
     * When a reply arrives this task must be canceled using the <code>cancel()</code>
     * method inherited from <code>TimerTask</code>. In this case the caller is
     * responsible for removing the task from the <code>tasks</code> map.
     */
    class TimeoutTask extends TimerTask {

        private final int comm;
        private final Receiver recv;

        public TimeoutTask(final int comm, final Receiver recv) {
            this.comm = comm;
            this.recv = recv;
        }

        @Override
        public void run() {
            Thread.currentThread().setName("TimoutTask " + KadServer.this.localNode);
            if (!KadServer.this.isRunning) {
                return;
            }

            try {
                unregister(this.comm);
                this.recv.timeout(this.comm);
            } catch (final IOException e) {
                System.err.println("Cannot unregister a receiver. Message: " + e.getMessage());
            }
        }
    }

    public void printReceivers() {
        for (final Integer r : this.receivers.keySet()) {
            System.out.println("Receiver for comm: " + r + "; Receiver: " + this.receivers.get(r));
        }
    }

    public boolean isRunning() {
        return this.isRunning;
    }

}
