package at.tugraz.iaik.p2p.kademlia;

import java.text.DecimalFormat;

/**
 * Class that keeps statistics for this Kademlia instance.
 * <p>
 * These statistics are temporary and will be lost when Kad is shut down.
 *
 * @author Joshua Kissoon
 * @since 20140505
 */
public class Statistician implements KadStatistician {

    /* How much data was sent and received by the server over the network */
    private long totalDataSent, totalDataReceived;
    private long numDataSent, numDataReceived;

    /* Bootstrap timings */
    private long bootstrapTime;

    /* Content lookup operation timing & route length */
    private int numContentLookups, numFailedContentLookups;
    private long totalContentLookupTime;
    private long totalRouteLength;


    {
        totalDataSent = 0;
        totalDataReceived = 0;
        bootstrapTime = 0;
        numContentLookups = 0;
        totalContentLookupTime = 0;
        totalRouteLength = 0;
    }

    @Override
    public synchronized void sentData(final long size) {
        totalDataSent += size;
        numDataSent++;
    }

    @Override
    public long getTotalDataSent() {
        if (totalDataSent == 0) {
            return 0L;
        }

        return totalDataSent / 1000L;
    }

    @Override
    public synchronized void receivedData(final long size) {
        totalDataReceived += size;
        numDataReceived++;
    }

    @Override
    public long getTotalDataReceived() {
        if (totalDataReceived == 0) {
            return 0L;
        }
        return totalDataReceived / 1000L;
    }

    @Override
    public void setBootstrapTime(final long time) {
        bootstrapTime = time;
    }

    @Override
    public long getBootstrapTime() {
        return bootstrapTime / 1000000L;
    }

    @Override
    public void addContentLookup(final long time, final int routeLength, final boolean isSuccessful) {
        if (isSuccessful) {
            numContentLookups++;
            totalContentLookupTime += time;
            totalRouteLength += routeLength;
        } else {
            numFailedContentLookups++;
        }
    }

    @Override
    public int numContentLookups() {
        return numContentLookups;
    }

    @Override
    public int numFailedContentLookups() {
        return numFailedContentLookups;
    }

    @Override
    public long totalContentLookupTime() {
        return totalContentLookupTime;
    }

    @Override
    public double averageContentLookupTime() {
        if (numContentLookups == 0) {
            return 0D;
        }

        final double        avg = (double) ((double) totalContentLookupTime / (double) numContentLookups) / 1000000D;
        final DecimalFormat df  = new DecimalFormat("#.00");
        return new Double(df.format(avg));
    }

    @Override
    public double averageContentLookupRouteLength() {
        if (numContentLookups == 0) {
            return 0D;
        }
        final double        avg = (double) ((double) totalRouteLength / (double) numContentLookups);
        final DecimalFormat df  = new DecimalFormat("#.00");
        return new Double(df.format(avg));
    }

    @Override
    public String toString() {
        final StringBuilder sb = new StringBuilder("Statistician: [");

        sb.append("Bootstrap Time: ");
        sb.append(getBootstrapTime());
        sb.append("; ");

        sb.append("Data Sent: ");
        sb.append("(");
        sb.append(numDataSent);
        sb.append(") ");
        sb.append(getTotalDataSent());
        sb.append(" keyBytes; ");

        sb.append("Data Received: ");
        sb.append("(");
        sb.append(numDataReceived);
        sb.append(") ");
        sb.append(getTotalDataReceived());
        sb.append(" keyBytes; ");

        sb.append("Num Content Lookups: ");
        sb.append(numContentLookups());
        sb.append("; ");

        sb.append("Avg Content Lookup Time: ");
        sb.append(averageContentLookupTime());
        sb.append("; ");

        sb.append("Avg Content Lookup Route Lth: ");
        sb.append(averageContentLookupRouteLength());
        sb.append("; ");

        sb.append("]");

        return sb.toString();
    }
}
