package at.tugraz.iaik.p2p.kademlia.message;

import at.tugraz.iaik.p2p.kademlia.KadConfiguration;
import at.tugraz.iaik.p2p.kademlia.KadServer;
import at.tugraz.iaik.p2p.kademlia.KademliaNode;
import at.tugraz.iaik.p2p.kademlia.node.Node;

import java.io.IOException;
import java.util.List;

/**
 * Receives a NodeLookupMessage and sends a NodeReplyMessage as reply with the K-Closest nodes to the ID sent.
 *
 * @author Joshua Kissoon
 * @created 20140219
 */
public class NodeLookupReceiver implements Receiver {

    private final KadServer server;
    private final KademliaNode localNode;
    private final KadConfiguration config;

    public NodeLookupReceiver(final KadServer server, final KademliaNode local, final KadConfiguration config) {
        this.server = server;
        this.localNode = local;
        this.config = config;
    }

    /**
     * Handle receiving a NodeLookupMessage
     * Find the set of K nodes closest to the lookup ID and return them
     *
     * @param comm
     * @throws java.io.IOException
     */
    @Override
    public void receive(final Message incoming, final int comm) throws IOException {
        final NodeLookupMessage msg = (NodeLookupMessage) incoming;

        final Node origin = msg.getOrigin();

        /* Update the local space by inserting the origin node. */
        this.localNode.getRoutingTable().insert(msg);

        /* Find nodes closest to the LookupId */
        final List<Message> nodes = this.localNode.getRoutingTable().findClosest(msg.getLookupId(), this.config.k());

        /* Respond to the NodeLookupMessage */
        final Message reply = new NodeReplyMessage(this.localNode.getNode(), origin, comm, nodes);

        if (this.server.isRunning()) {
            /* Let the Server send the reply */
            this.server.reply(reply);
        }
    }

    /**
     * We don't need to do anything here
     *
     * @param comm
     * @throws java.io.IOException
     */
    @Override
    public void timeout(final int comm) throws IOException {
    }
}
