/**
 * @author Joshua Kissoon
 * @created 20140218
 * @desc Operation that handles connecting to an existing Kademlia network using a bootstrap node
 */
package at.tugraz.iaik.p2p.kademlia.operation;

import at.tugraz.iaik.p2p.kademlia.KadConfiguration;
import at.tugraz.iaik.p2p.kademlia.KadServer;
import at.tugraz.iaik.p2p.kademlia.KademliaNode;
import at.tugraz.iaik.p2p.kademlia.exceptions.RoutingException;
import at.tugraz.iaik.p2p.kademlia.message.Message;
import at.tugraz.iaik.p2p.kademlia.message.NodeLookupMessage;
import at.tugraz.iaik.p2p.kademlia.message.NodeReplyMessage;
import at.tugraz.iaik.p2p.kademlia.message.Receiver;
import at.tugraz.iaik.p2p.kademlia.node.Node;

import java.io.IOException;

public class ConnectOperation implements Operation, Receiver {

    public static final int MAX_CONNECT_ATTEMPTS = 15;       // Try 5 times to connect to a node

    private final KadServer server;
    private final KademliaNode localNode;
    private final Node bootstrapNode;
    private final KadConfiguration config;

    private boolean error;
    private int attempts;

    /**
     * @param server    The message server used to send/receive messages
     * @param local     The local node
     * @param bootstrap Node to use to bootstrap the local node onto the network
     * @param config
     */
    public ConnectOperation(final KadServer server, final KademliaNode local, final Node bootstrap, final
    KadConfiguration config) {
        this.server = server;
        this.localNode = local;
        this.bootstrapNode = bootstrap;
        this.config = config;
    }

    @Override
    public synchronized void execute() throws IOException {
        try {
            /* Contact the bootstrap node */
            this.error = true;
            this.attempts = 0;
            final Message m = new NodeLookupMessage(localNode.getNode(), bootstrapNode, localNode.getNode().getNodeId
                    (), server.random.nextInt());

            /* Send a connect message to the bootstrap node */
            this.server.sendMessage(m, this);

            /* If we haven't finished as yet, wait for a maximum of config.operationTimeout() time */
            int totalTimeWaited = 0;
            final int timeInterval = 100;     // We re-check every 300 milliseconds
            while (totalTimeWaited < this.config.operationTimeout()) {
                if (this.error) {
                    wait(timeInterval);
                    totalTimeWaited += timeInterval;
                } else {
                    break;
                }
            }
            if (this.error) {
                /* If we still haven't received any responses by then, do a routing timeout */
                throw new RoutingException("ConnectOperation: Bootstrap node did not respond: " + this.bootstrapNode);
            }

//            /* Perform lookup for our own ID to get nodes close to us */
//            final Operation lookup = new NodeLookupOperation(this.server, this.localNode, this.localNode.getMessage()
//                    .getNodeId(), this.config);
//            lookup.execute();

            /**
             * Refresh buckets to get a good routing table
             * After the above lookup operation, K nodes will be in our routing table,
             * Now we try to populate all of our buckets.
             */
            new BucketRefreshOperation(this.server, this.localNode, this.config).execute();
        } catch (final InterruptedException e) {
            System.err.println("Connect operation was interrupted. ");
        }
    }

    /**
     * Receives an AcknowledgeMessage from the bootstrap node.
     *
     * @param comm
     */
    @Override
    public synchronized void receive(final Message incoming, final int comm) {
        /* The incoming message will be an acknowledgement message */
        if (!(incoming instanceof NodeReplyMessage)) return;


        /* The bootstrap node has responded, insert it into our space */
        NodeReplyMessage reply = (NodeReplyMessage) incoming;
        this.localNode.getRoutingTable().insert(reply.getOriginMessage());
        for (Message msg : reply.getNodes()) {
            localNode.getRoutingTable().insert(msg);
        }

        /* We got a response, so the error is false */
        this.error = false;

        /* Wake up any waiting thread */
        notify();
    }

    /**
     * Resends a ConnectMessage to the boot strap node a maximum of MAX_ATTEMPTS
     * times.
     *
     * @param comm
     * @throws java.io.IOException
     */
    @Override
    public synchronized void timeout(final int comm) throws IOException {
        if (++this.attempts < MAX_CONNECT_ATTEMPTS) {
            this.server.sendMessage(new NodeLookupMessage(localNode.getNode(), bootstrapNode, localNode.getNode()
                    .getNodeId(), server.random.nextInt()));
        } else {
            /* We just exit, so notify all other threads that are possibly waiting */
            notify();
        }
    }
}
