package at.tugraz.iaik.p2p.kademlia.operation;

import at.tugraz.iaik.p2p.kademlia.KadConfiguration;
import at.tugraz.iaik.p2p.kademlia.exceptions.RoutingException;
import at.tugraz.iaik.p2p.kademlia.node.Identifier;
import at.tugraz.iaik.p2p.kademlia.KadServer;
import at.tugraz.iaik.p2p.kademlia.KademliaNode;
import at.tugraz.iaik.p2p.kademlia.message.Message;
import at.tugraz.iaik.p2p.kademlia.message.NodeLookupMessage;
import at.tugraz.iaik.p2p.kademlia.message.NodeReplyMessage;
import at.tugraz.iaik.p2p.kademlia.message.Receiver;
import at.tugraz.iaik.p2p.kademlia.node.Node;
import at.tugraz.iaik.p2p.kademlia.node.NodeKeyComparator;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

/**
 * Finds the K closest nodes to a specified identifier
 * The algorithm terminates when it has gotten responses from the K closest nodes it has seen.
 * Nodes that fail to respond are removed from consideration
 *
 * @author Joshua Kissoon
 * @created 20140219
 */
public class NodeLookupOperation implements Operation, Receiver {

    /* Constants */
    private static final String UNASKED = "UnAsked";
    private static final String AWAITING = "Awaiting";
    private static final String ASKED = "Asked";
    private static final String FAILED = "Failed";

    private final KadServer server;
    private final KademliaNode localNode;
    private final KadConfiguration config;

    private final Identifier lookupId;        // query
    private final Map<Node, String> nodes;

    /* Tracks messages in transit and awaiting reply */
    private final Map<Integer, Node> messagesTransiting;

    /* Used to sort nodes */
    private final NodeKeyComparator comparator;


    {
        this.messagesTransiting = new HashMap<>();
    }

    /**
     * @param server    KadServer used for communication
     * @param localNode The local node making the communication
     * @param lookupId  The ID for which to find nodes close to
     * @param config
     */
    public NodeLookupOperation(final KadServer server, final KademliaNode localNode, final Identifier lookupId, final
    KadConfiguration config) {
        this.server = server;
        this.localNode = localNode;
        this.config = config;

        this.lookupId = lookupId;

        /**
         * We initialize a TreeMap to store nodes.
         * This map will be sorted by which nodes are closest to the lookupId
         */
        this.comparator = new NodeKeyComparator(lookupId);
        this.nodes = new TreeMap(this.comparator);
    }

    /**
     * @throws java.io.IOException
     * @throws RoutingException
     */
    @Override
    public synchronized void execute() throws IOException, RoutingException {
        try {
            /* Set the local node as already asked */
            this.nodes.put(localNode.getNode(), ASKED);

            /**
             * We add all nodes here instead of the K-Closest because there may be the case that the K-Closest are
             * offline
             * - The operation takes care of looking at the K-Closest.
             */
            this.addNodes(this.localNode.getRoutingTable().getAllNodes());

            /* If we haven't finished as yet, wait for a maximum of config.operationTimeout() time */
            int totalTimeWaited = 0;
            final int timeInterval = 10;     // We re-check every n milliseconds
            while (totalTimeWaited < this.config.operationTimeout()) {
                if (!this.askNodesorFinish()) {
                    wait(timeInterval);
                    totalTimeWaited += timeInterval;
                } else {
                    break;
                }
            }

            /* Now after we've finished, we would have an idea of offline nodes, lets update our routing table */
            this.localNode.getRoutingTable().setUnresponsiveContacts(this.getFailedNodes());

        } catch (final InterruptedException e) {
            throw new RuntimeException(e);
        }
    }

    public List<Node> getClosestNodes() {
        return this.closestNodes(ASKED);
    }

    /**
     * Add nodes from this list to the set of nodes to lookup
     *
     * @param list The list from which to add nodes
     */
    private void addNodes(final List<Message> list) {
        for (final Message o : list) {
            /* If this node is not in the list, add the node */
            if (!this.nodes.containsKey(o.getOrigin())) {
                this.nodes.put(o.getOrigin(), UNASKED);
            }
        }
    }

    /**
     * Asks some of the K closest nodes seen but not yet queried.
     * Assures that no more than DefaultConfiguration.CONCURRENCY messages are in transit at a time
     * <p>
     * This method should be called every time a reply is received or a timeout occurs.
     * <p>
     * If all K closest nodes have been asked and there are no messages in transit,
     * the algorithm is finished.
     *
     * @return <code>true</code> if finished OR <code>false</code> otherwise
     */
    private boolean askNodesorFinish() throws IOException {
        /* If >= CONCURRENCY nodes are in transit, don't do anything */
        if (this.config.maxConcurrentMessagesTransiting() <= this.messagesTransiting.size()) {
            return false;
        }

        /* Get unqueried nodes among the K closest seen that have not FAILED */
        final List<Node> unasked = this.closestNodesNotFailed(UNASKED);

        if (unasked.isEmpty() && this.messagesTransiting.isEmpty()) {
            /* We have no unasked nodes nor any messages in transit, we're finished! */
            return true;
        }

        /**
         * Send messages to nodes in the list;
         * making sure than no more than CONCURRENCY messsages are in transit
         */
        for (int i = 0; (this.messagesTransiting.size() < this.config.maxConcurrentMessagesTransiting()) && (i <
                unasked.size()); i++) {
            final Node n = (Node) unasked.get(i);

            final int comm = this.server.sendMessage(new NodeLookupMessage(localNode.getNode(), n, lookupId, server
                    .random.nextInt()), this);

            this.nodes.put(n, AWAITING);
            this.messagesTransiting.put(comm, n);
        }

        /* We're not finished as yet, return false */
        return false;
    }

    /**
     * @param status The status of the nodes to return
     * @return The K closest nodes to the target lookupId given that have the specified status
     */
    private List<Node> closestNodes(final String status) {
        final List<Node> closestNodes = new ArrayList<>(this.config.k());
        int remainingSpaces = this.config.k();

        for (final Map.Entry e : this.nodes.entrySet()) {
            if (status.equals(e.getValue())) {
                /* We got one with the required status, now add it */
                closestNodes.add((Node) e.getKey());
                if (--remainingSpaces == 0) {
                    break;
                }
            }
        }

        return closestNodes;
    }

    /**
     * Find The K closest nodes to the target lookupId given that have not FAILED.
     * From those K, get those that have the specified status
     *
     * @param status The status of the nodes to return
     * @return A List of the closest nodes
     */
    private List<Node> closestNodesNotFailed(final String status) {
        final List<Node> closestNodes = new ArrayList<>(this.config.k());
        int remainingSpaces = this.config.k();

        for (final Map.Entry<Node, String> e : this.nodes.entrySet()) {
            if (!FAILED.equals(e.getValue())) {
                if (status.equals(e.getValue())) {
                    /* We got one with the required status, now add it */
                    closestNodes.add(e.getKey());
                }

                if (--remainingSpaces == 0) {
                    break;
                }
            }
        }

        return closestNodes;
    }

    /**
     * Receive and handle the incoming NodeReplyMessage
     *
     * @param comm
     * @throws java.io.IOException
     */
    @Override
    public synchronized void receive(final Message incoming, final int comm) throws IOException {
        if (!(incoming instanceof NodeReplyMessage)) {
            /* Not sure why we get a message of a different type here... @todo Figure it out. */
            return;
        }
        /* We receive a NodeReplyMessage with a set of nodes, read this message */
        final NodeReplyMessage msg = (NodeReplyMessage) incoming;

        /* Add the origin node to our routing table */
        final Node origin = msg.getOrigin();
        this.localNode.getRoutingTable().insert(msg.getOriginMessage());

        /* Set that we've completed ASKing the origin node */
        this.nodes.put(origin, ASKED);

        /* Remove this msg from messagesTransiting since it's completed now */
        this.messagesTransiting.remove(comm);

        /* Add the received nodes to our nodes list to query */
        this.addNodes(msg.getNodes());
        this.askNodesorFinish();
    }

    /**
     * A node does not respond or a packet was lost, we set this node as failed
     *
     * @param comm
     * @throws java.io.IOException
     */
    @Override
    public synchronized void timeout(final int comm) throws IOException {
        /* Get the node associated with this communication */
        final Node n = this.messagesTransiting.get(comm);

        if (n == null) {
            return;
        }

        /* Mark this node as failed and inform the routing table that it is unresponsive */
        this.nodes.put(n, FAILED);
        this.localNode.getRoutingTable().setUnresponsiveContact(n);
        this.messagesTransiting.remove(comm);

        this.askNodesorFinish();
    }

    public List<Node> getFailedNodes() {
        final List<Node> failedNodes = new ArrayList<>();

        for (final Map.Entry<Node, String> e : this.nodes.entrySet()) {
            if (e.getValue().equals(FAILED)) {
                failedNodes.add(e.getKey());
            }
        }

        return failedNodes;
    }
}
