package at.tugraz.iaik.p2p.kademlia.routing;

import at.tugraz.iaik.p2p.kademlia.KadConfiguration;
import at.tugraz.iaik.p2p.kademlia.message.Message;
import at.tugraz.iaik.p2p.kademlia.message.SimpleMessage;
import at.tugraz.iaik.p2p.kademlia.node.Identifier;
import at.tugraz.iaik.p2p.kademlia.node.KeyComparator;
import at.tugraz.iaik.p2p.kademlia.node.Node;
import at.tugraz.iaik.p2p.kademlia.node.SHA3ID;

import java.util.*;

/**
 * Implementation of a Kademlia routing table
 *
 * @author Joshua Kissoon
 * @created 20140215
 */
public class JKademliaRoutingTable implements KademliaRoutingTable {


    private final BackRefRoutingTable backrefMap = new BackRefRoutingTable();

    private final     Node             localNode;  // The current node
    private transient KademliaBucket[] buckets;

    private transient KadConfiguration config;

    public JKademliaRoutingTable(final Node localNode, final KadConfiguration config) {
        this.localNode = localNode;
        this.config = config;

        /* Initialize all of the buckets to a specific depth */
        this.initialize();

        /* Insert the local node */
        this.insert(new SimpleMessage(localNode, localNode, 0, ""));
    }

    /**
     * Initialize the JKademliaRoutingTable to it's default state
     */
    @Override
    public final void initialize() {
        this.buckets = new KademliaBucket[this.localNode.getNodeId().getNumBits()];
        for (int i = 0; i < this.localNode.getNodeId().getNumBits(); i++) {
            this.buckets[i] = new JKademliaBucket(i, this.config);
        }
    }

    @Override
    public void setConfiguration(final KadConfiguration config) {
        this.config = config;
    }

    /**
     * Adds a contact to the routing table based on how far it is from the LocalNode.
     *
     * @param c The contact to add
     */
    @Override
    public synchronized final void insert(final Contact c) {
        Contact stalest = this.buckets[this.getBucketId(c.getMessage().getOrigin().getNodeId())].insert(c);
        if (stalest != null) {
            backrefMap.remove(stalest.getMessage().getOrigin());
        }
        backrefMap.add(c.getMessage().getOrigin());
    }

    /**
     * Adds a node to the routing table based on how far it is from the LocalNode.
     *
     * @param n The node to add
     */
    @Override
    public synchronized final void insert(final Message n) {
        Contact stalest = this.buckets[this.getBucketId(n.getOrigin().getNodeId())].insert(n);
        if (stalest != null) {
            backrefMap.remove(stalest.getMessage().getOrigin());
        }
        backrefMap.add(n.getOrigin());
    }

    /**
     * Compute the bucket ID in which a given node should be placed; the bucketId is computed based on how far the
     * node is away from the Local Node.
     *
     * @param nid The NodeId for which we want to find which bucket it belong to
     * @return Integer The bucket ID in which the given node should be placed.
     */
    @Override
    public final int getBucketId(final Identifier nid) {
        final int bId = this.localNode.getNodeId().getDistance(nid) - 1;

        /* If we are trying to insert a node into it's own routing table, then the bucket ID will be -1, so let's
        just keep it in bucket 0 */
        return bId < 0 ? 0 : bId;
    }

    /**
     * Find the closest set of contacts to a given NodeId
     *
     * @param target           The NodeId to find contacts close to
     * @param numNodesRequired The number of contacts to find
     * @return List A List of contacts closest to target
     */
    @Override
    public synchronized final List<Message> findClosest(final Identifier target, final int numNodesRequired) {
        final TreeSet<Message> sortedSet = new TreeSet<Message>(new KeyComparator(target));
        sortedSet.addAll(this.getAllNodes());

        final List<Message> closest = new ArrayList<>(numNodesRequired);

        /* Now we have the sorted set, lets get the top numRequired */
        int count = 0;
        for (final Message n : sortedSet) {
            closest.add(n);
            if (++count == numNodesRequired) {
                break;
            }
        }
        return closest;
    }

    @Override
    public synchronized final List<Message> findClosestforData(final SHA3ID target, final int numNodesRequired) {
        Set<Node>    closestIDs   = backrefMap.getClosestNodes(target, numNodesRequired);
        ArrayList<Message> closestNodes = new ArrayList<>(closestIDs.size());

        List<Message> allNodes = getAllNodes();
        for (Message m : allNodes) {
            for (Node i : closestIDs) {
                if (m.getOrigin().equals(i)) {
                    closestNodes.add(m);
                }
            }
        }
        return closestNodes;
    }

    /**
     * @return List A List of all Nodes in this JKademliaRoutingTable
     */
    @Override
    public synchronized final List<Message> getAllNodes() {
        final List<Message> nodes = new ArrayList<>();

        for (final KademliaBucket b : this.buckets) {
            for (final Contact c : b.getContacts()) {
                nodes.add(c.getMessage());
            }
        }

        return nodes;
    }

    /**
     * @return List A List of all Nodes in this JKademliaRoutingTable
     */
    @Override
    public final List<Contact> getAllContacts() {
        final List<Contact> contacts = new ArrayList<>();

        for (final KademliaBucket b : this.buckets) {
            contacts.addAll(b.getContacts());
        }

        return contacts;
    }

    /**
     * @return Bucket[] The buckets in this Kad Instance
     */
    @Override
    public final KademliaBucket[] getBuckets() {
        return this.buckets;
    }

    /**
     * Set the KadBuckets of this routing table, mainly used when retrieving saved state
     *
     * @param buckets
     */
    public final void setBuckets(final KademliaBucket[] buckets) {
        this.buckets = buckets;
    }

    /**
     * Method used by operations to notify the routing table of any contacts that have been unresponsive.
     *
     * @param contacts The set of unresponsive contacts
     */
    @Override
    public void setUnresponsiveContacts(final List<Node> contacts) {
        if (contacts.isEmpty()) {
            return;
        }
        for (final Node n : contacts) {
            this.setUnresponsiveContact(n);
        }
    }

    /**
     * Method used by operations to notify the routing table of any contacts that have been unresponsive.
     *
     * @param n
     */
    @Override
    public synchronized void setUnresponsiveContact(final Node n) {
        final int bucketId = this.getBucketId(n.getNodeId());

        /* Remove the contact from the bucket */
        this.buckets[bucketId].removeNode(n);
    }

    @Override
    public synchronized final String toString() {
        final StringBuilder sb            = new StringBuilder("\nPrinting Routing Table Started ***************** \n");
        int                 totalContacts = 0;
        sb.append(localNode.getNodeId()).append("\n");
        for (final KademliaBucket b : this.buckets) {
            if (b.numContacts() > 0) {
                totalContacts += b.numContacts();
                sb.append("# nodes in Bucket with depth ");
                sb.append(b.getDepth());
                sb.append(": ");
                sb.append(b.numContacts());
                sb.append("\n");
                sb.append(b.toString());
                sb.append("\n");
            }
        }

        sb.append("\nTotal Contacts: ");
        sb.append(totalContacts);
        sb.append("\n\n");

        sb.append("Data Routing Table\n\n");
        for (Map.Entry<Node, Set<Identifier>> e : backrefMap.getSnapshot().entrySet()) {
            sb.append(e.getKey()).append(": ");
            for(Identifier i : e.getValue())
            sb.append(i).append(", ");
            sb.append("\n");
        }
        sb.append("\n");
        sb.append("Printing Routing Table Ended ******************** ");

        return sb.toString();
    }

}
