package at.tugraz.iaik.p2p.kademlia.util.serializer;

import at.tugraz.iaik.p2p.kademlia.KadConfiguration;
import at.tugraz.iaik.p2p.kademlia.routing.Contact;
import at.tugraz.iaik.p2p.kademlia.routing.JKademliaRoutingTable;
import at.tugraz.iaik.p2p.kademlia.routing.KademliaRoutingTable;
import com.google.gson.Gson;
import com.google.gson.reflect.TypeToken;
import com.google.gson.stream.JsonReader;
import com.google.gson.stream.JsonWriter;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.lang.reflect.Type;
import java.util.List;

/**
 * A KadSerializer that serializes routing tables to JSON format
 * The generic serializer is not working for routing tables
 * <p>
 * Why a JKademliaRoutingTable specific serializer?
 * The routing table structure:
 * - JKademliaRoutingTable
 * -- Buckets[]
 * --- Map<NodeId, Node>
 * ---- NodeId:KeyBytes
 * ---- Node: NodeId, InetAddress, Port
 * <p>
 * The above structure seems to be causing some problem for Gson,
 * especially at the Map part.
 * <p>
 * Solution
 * - Make the Buckets[] transient
 * - Simply store all Nodes in the serialized object
 * - When reloading, re-add all nodes to the JKademliaRoutingTable
 *
 * @author Joshua Kissoon
 * @since 20140310
 */
public class JsonRoutingTableSerializer implements KadSerializer<KademliaRoutingTable> {

    private final Gson gson;

    Type contactCollectionType = new TypeToken<List<Contact>>() {
    }.getType();

    private final KadConfiguration config;


    {
        this.gson = new Gson();
    }

    /**
     * Initialize the class
     *
     * @param config
     */
    public JsonRoutingTableSerializer(final KadConfiguration config) {
        this.config = config;
    }

    @Override
    public void write(final KademliaRoutingTable data, final DataOutputStream out) throws IOException {
        try (JsonWriter writer = new JsonWriter(new OutputStreamWriter(out))) {
            writer.beginArray();

            /* Write the basic JKademliaRoutingTable */
            this.gson.toJson(data, JKademliaRoutingTable.class, writer);

            /* Now Store the Contacts  */
            this.gson.toJson(data.getAllContacts(), this.contactCollectionType, writer);

            writer.endArray();
        }
    }

    @Override
    public KademliaRoutingTable read(final DataInputStream in) throws IOException, ClassNotFoundException {
        try (DataInputStream din = new DataInputStream(in); JsonReader reader = new JsonReader(new InputStreamReader
                (in))) {
            reader.beginArray();

            /* Read the basic JKademliaRoutingTable */
            final KademliaRoutingTable tbl = this.gson.fromJson(reader, KademliaRoutingTable.class);
            tbl.setConfiguration(this.config);
            
            /* Now get the Contacts and add them back to the JKademliaRoutingTable */
            final List<Contact> contacts = this.gson.fromJson(reader, this.contactCollectionType);
            tbl.initialize();

            for (final Contact c : contacts) {
                tbl.insert(c);
            }

            reader.endArray();
            /* Read and return the Content*/
            return tbl;
        }
    }
}
