package at.tugraz.iaik.p2p.kademlia.net

import at.tugraz.iaik.p2p.kademlia.KadStatistician
import at.tugraz.iaik.p2p.kademlia.message.Message
import at.tugraz.iaik.p2p.kademlia.message.MessageFactory
import java.io.*
import java.net.DatagramPacket
import java.net.DatagramSocket
import java.net.InetAddress
import java.net.InetSocketAddress

/**
 * Created by Bernd Prünster <bernd.pruenster></bernd.pruenster>@iaik.tugraz.at>
 */
open class UdpConnector protected constructor(type: Connector.Type,
                                              val address: InetAddress?,
                                              val port: Int,
                                              statistician: KadStatistician? = null) : Connector(type, statistician) {
    override val destinationKnown: Boolean
        get() = address != null

    @JvmOverloads constructor(address: InetAddress? = InetAddress.getLoopbackAddress(),
                              port: Int,
                              statistician: KadStatistician? = null) : this(Connector.Type.UDP,
                                                                            address,
                                                                            port,
                                                                            statistician)

    private var socket: DatagramSocket? = null


    @Throws(IOException::class)
    override fun toStream(out: DataOutputStream) {
        if (address == null) out.writeByte(0) else {
            out.writeByte(address.address.size)
            out.write(address.address)
        }
        out.writeInt(this.port)
    }


    @Throws(IOException::class)
    override fun receive(): MessageMapping {

        /* Wait for a packet */

        val buffer = ByteArray(DATAGRAM_BUFFER_SIZE)
        val packet = DatagramPacket(buffer, buffer.size)
        var received = false;
        socket?.apply {
            receive(packet)
            received = true
        }
        if (!received) throw IOException()


        /* Lets inform the statistician that we've received some data */
        statistician?.receivedData(packet.length.toLong())


        /* We've received a packet, now handle it */
        val parsed = parsePacket(packet)
        ByteArrayInputStream(parsed).use { bin ->
            DataInputStream(bin).use { din ->
                /* Read in the conversation Id to know which handler to handle this response */
                val msg = MessageFactory.createMessage(din)
                din.close()
                return MessageMapping(msg.commID, msg)
            }
        }
    }

    @Throws(IOException::class)
    protected open fun parsePacket(packet: DatagramPacket): ByteArray {
        val offset = packet.offset
        val result = ByteArray(packet.length - offset)
        System.arraycopy(packet.data, offset, result, 0, packet.length - offset)
        return result
    }

    @Throws(IOException::class)
    override fun send(msg: Message) {
        val destination = msg.target.connector
        if (destination !is UdpConnector) throw IOException("Cannot send message to " + destination.javaClass.simpleName + " but only" + " to " + javaClass.simpleName)


        val data = serialiseMessage(msg)

        if (data.size > DATAGRAM_BUFFER_SIZE) {
            throw IOException("Message is too big")
        }

        /* Everything is good, now create the packet and send it */
        val pkt = DatagramPacket(data, 0, data.size)
        pkt.socketAddress = InetSocketAddress(destination.address, destination.port)
        var sent = false
        socket?.apply {
            send(pkt)
            sent = true
        }
        if (!sent) throw IOException("socket was closed propr to sending")

        /* Lets inform the statistician that we've sent some data */
        statistician?.sentData(data.size.toLong())

    }

    @Throws(IOException::class)
    protected open fun serialiseMessage(msg: Message): ByteArray {
        ByteArrayOutputStream().use { bout ->
            DataOutputStream(bout).use { dout ->
                /* Setup the message for transmission */
                // dout.writeByte(msg.code());
                msg.toStream(dout)
                dout.close()

                return bout.toByteArray()
            }
        }
    }


    @Throws(IOException::class)
    override fun open() {
        synchronized(this) {
            socket?.let { throw IOException("Already open!") }
            socket = DatagramSocket(this.port)
        }
    }


    @Throws(IOException::class)
    override fun close() {
        synchronized(this) {
            socket?.let {
                socket?.close()
                socket = null
            }
        }
    }

    companion object {
        /* Maximum size of a Datagram Packet */
        private val DATAGRAM_BUFFER_SIZE = 64 * 1024      // 64KB
    }
}
