package at.tugraz.iaik.p2p.kademlia.node

import at.tugraz.iaik.p2p.kademlia.message.OutStreamable
import java.io.DataInputStream
import java.io.IOException
import java.io.Serializable
import java.math.BigInteger
import java.util.*
import kotlin.experimental.xor

abstract class Identifier(val keyBytes: ByteArray) : OutStreamable, Serializable {

    abstract val type: Type

    abstract val numBits: Int

    /**
     * @return The BigInteger representation of the key
     */
    val int: BigInteger
        get() = BigInteger(1, keyBytes)

    enum class Type constructor(val id: Byte) {
        KAD(1.toByte()),
        EC(2.toByte()),
        SHA3(3.toByte());


        companion object {

            fun fromID(id: Byte): Type {
                for (t in Type.values()) {
                    if (id == t.id) {
                        return t
                    }
                }
                throw IllegalArgumentException("Unknown type: " + id)
            }
        }
    }

    init {
        validate(keyBytes)
    }

    abstract fun generateNodeIdByDistance(distance: Int): Identifier

    protected abstract fun validate(bytes: ByteArray)

    /**
     * Checks the distance between this and another NodeId
     *
     * @param nid
     * @return The distance of this NodeId from the given NodeId
     */
    protected abstract fun xor(nid: Identifier): Identifier

    /**
     * Gets the distance from this NodeId to another NodeId
     *
     * @param to
     * @return Integer The distance
     */
    abstract fun getDistance(to: Identifier): Int

    fun hexRepresentation(): String {
        /* Returns the hex format of this NodeId */
        val bi = BigInteger(1, keyBytes)
        return String.format("%0" + (keyBytes.size shl 1) + "X", bi)
    }

    override fun toString(): String {
        return hexRepresentation()
    }

    /**
     * Counts the number of leading 0's in this NodeId
     *
     * @return Integer The number of leading 0's
     */
    /* If the byte is not 0, we need to count how many MSBs are 0 */// Reset the count if we encounter a non-zero number
    /* Add the count of MSB 0s to the prefix length *//* Break here since we've now covered the MSB 0s */
    fun firstSetBitIndex(offset: Int = 0): Int {
        var off = offset
        var prefixLength = 0

        for (b in keyBytes) {
            val keyByte = b.toInt()
            if (off > 0) {
                off -= 1
                prefixLength += 8
            } else if (keyByte == 0) {
                prefixLength += 8
            } else {
                var count = 0
                for (i in 7 downTo 0) {
                    val a = keyByte and (1 shl i) == 0
                    if (a) {
                        count++
                    } else {
                        break
                    }
                }
                prefixLength += count
                break
            }
        }
        return prefixLength
    }

    override fun equals(other: Any?): Boolean {
        if (other !is Identifier) return false
        return Arrays.equals(keyBytes, other.keyBytes)
    }

    companion object {

        @JvmStatic
        fun forData(idType: Type, bytes: ByteArray): Identifier {
            return when (idType) {
                Identifier.Type.KAD  -> KademliaID(bytes)
                Identifier.Type.EC   -> KeyBasedID(bytes)
                Identifier.Type.SHA3 -> SHA3ID(bytes)
            }
        }

        @Throws(IOException::class)
        @JvmStatic
        fun fromStr(`in`: DataInputStream): Identifier {
            val type = Identifier.Type.fromID(`in`.readByte())
            return when (type) {
                Identifier.Type.KAD  -> KademliaID.fromStr(`in`)
                Identifier.Type.EC   -> KeyBasedID.fromStr(`in`)
                Identifier.Type.SHA3 -> SHA3ID.fromStr(`in`)
            }
        }

        @JvmStatic
        @JvmOverloads
        fun xor(a: ByteArray, b: ByteArray, offset: Int = 0): ByteArray {
            if (a.size != b.size) {
                throw IllegalArgumentException("array lengths are not equal")
            }
            val result = ByteArray(a.size)
            for (i in 0 until offset) {
                result[i] = 0
            }
            for (i in offset until result.size) {
                result[i] = (a[i] xor b[i])
            }

            return result
        }
    }

}