package at.tugraz.iaik.p2p.kademlia.node

import java.io.DataInputStream
import java.io.DataOutputStream
import java.io.IOException
import java.util.*

open class KademliaID(bytes: ByteArray) : Identifier(bytes) {

    /**
     * Construct the NodeId from some string
     *
     * @param data The user generated key string
     */
    constructor(data: String) : this(data.toByteArray())

    override val type = Identifier.Type.KAD

    override val numBits = ID_LENGTH

    open override fun validate(bytes: ByteArray) {
        if (bytes.size != ID_LENGTH / 8) {
            throw IllegalArgumentException("Specified Data need to be " + ID_LENGTH / 8 + " characters long. " + "Data Given: '" + String(
                    bytes) + "'")

        }
    }

    override fun equals(other: Any?): Boolean {
        return if (other is KademliaID) {
            keyBytes contentEquals other.keyBytes
        } else false
    }

    override fun hashCode(): Int {
        var hash = 7
        hash = 83 * hash + keyBytes.contentHashCode()
        return hash
    }

    /**
     * Checks the distance between this and another NodeId
     *
     * @param nid
     * @return The distance of this NodeId from the given NodeId
     */
    public override fun xor(nid: Identifier): KademliaID {
        return KademliaID(xor(this.keyBytes, nid.keyBytes))
    }

    /**
     * Generates a NodeId that is some distance away from this NodeId
     *
     * @param distance in number of bits
     * @return NodeId The newly generated NodeId
     */
    override fun generateNodeIdByDistance(distance: Int): KademliaID {
        val result = ByteArray(ID_LENGTH / 8)

        /* Since distance = ID_LENGTH - prefixLength, we need to fill that amount with 0's */
        val numByteZeroes = (ID_LENGTH - distance) / 8
        val numBitZeroes = 8 - distance % 8

        /* Filling byte zeroes */
        for (i in 0 until numByteZeroes) {
            result[i] = 0
        }

        /* Filling bit zeroes */
        val bits = BitSet(8)
        bits.set(0, 8)

        for (i in 0 until numBitZeroes) {
            /* Shift 1 zero into the start of the value */
            bits.clear(i)
        }
        bits.flip(0, 8)        // Flip the bits since they're in reverse order
        result[numByteZeroes] = bits.toByteArray()[0].toByte()

        /* Set the remaining keyBytes to Maximum value */
        for (i in numByteZeroes + 1 until result.size) {
            result[i] = java.lang.Byte.MAX_VALUE
        }

        return this.xor(KademliaID(result))
    }

    /**
     * Gets the distance from this NodeId to another NodeId
     *
     * @param to
     * @return Integer The distance
     */
    override fun getDistance(to: Identifier): Int {
        /**
         * Compute the xor of this and to
         * Get the index i of the first set bit of the xor returned NodeId
         * The distance between them is ID_LENGTH - i
         */
        return ID_LENGTH - this.xor(to).firstSetBitIndex()
    }

    @Throws(IOException::class)
    override fun toStream(out: DataOutputStream) {
        out.writeByte(type.id.toInt())
        out.write(this.keyBytes)
    }

    companion object {

        @Transient
        private val ID_LENGTH = 160


        fun random(): ByteArray {
            val keyBytes = ByteArray(ID_LENGTH / 8)
            Random().nextBytes(keyBytes)
            return keyBytes
        }

        @Throws(IOException::class)
        fun fromStr(`in`: DataInputStream): KademliaID {
            val input = ByteArray(ID_LENGTH / 8)
            `in`.readFully(input)
            return KademliaID(input)
        }
    }


}