package at.tugraz.iaik.p2p.kademlia.node

import at.tugraz.iaik.p2p.commons.CommonCrypto
import at.tugraz.iaik.p2p.commons.CryptoException
import java.io.DataInputStream
import java.io.DataOutputStream
import java.io.IOException
import java.security.interfaces.ECPublicKey

class KeyBasedID
/**
 * Generate the NodeId from a given byte[]
 *
 * @param bytes
 */
@Throws(CryptoException::class) private constructor(bytes: ByteArray, hash: Boolean, val key: ECPublicKey?) : SHA3ID(
        idFromInput(bytes, hash)) {


    @Throws(CryptoException::class) constructor(data: ByteArray) : this(data, true, CommonCrypto.decode(data))

    @Throws(CryptoException::class) constructor(key: ECPublicKey) : this(CommonCrypto.encode(key), true, key)

    @Throws(CryptoException::class) private constructor(encoded: Array<ByteArray?>) : this(encoded[0]!!,
                                                                                           false,
                                                                                           if (encoded[1] != null) CommonCrypto.decode(
                                                                                                   encoded[1]!!) else null)

    override val type = Identifier.Type.EC


    /**
     * Checks the distance between this and another NodeId
     *
     * @param nid
     * @return The distance of this NodeId from the given NodeId
     */
    override fun xor(nid: Identifier): KeyBasedID {
        try {
            return KeyBasedID(Identifier.xor(this.keyBytes, nid.keyBytes, 1), false, null)
        } catch (e: CryptoException) {
            throw RuntimeException(e)
        }

    }

    /**
     * Generates a NodeId that is some distance away from this NodeId
     *
     * @param distance in number of bits
     * @return NodeId The newly generated NodeId
     */
    override fun generateNodeIdByDistance(distance: Int): KeyBasedID {
        return xor(KeyBasedID(generateBytesByDitance(distance), false, null))
    }

    /**
     * Gets the distance from this NodeId to another NodeId
     *
     * @param to
     * @return Integer The distance
     */
    override fun getDistance(to: Identifier): Int {
        /**
         * Compute the xor of this and to
         * Get the index i of the first set bit of the xor returned NodeId
         * The distance between them is ID_LENGTH - i
         */
        return ID_LENGTH - xor(to).firstSetBitIndex(1)
    }

    @Throws(IOException::class)
    override fun toStream(out: DataOutputStream) {
        out.writeByte(type.id.toInt())
        out.writeBoolean(key != null)
        try {
            if (key != null) {
                out.write(CommonCrypto.encode(key))
            } else {
                out.write(this.keyBytes)
            }
        } catch (e: Exception) {
            throw IOException(e)
        }

    }

    companion object {

        @Throws(CryptoException::class)
        private fun idFromInput(data: ByteArray, hash: Boolean): ByteArray {
            if (!hash) {
                return data
            }
            return CommonCrypto.sha3(data)
        }

        // @Override
        @Throws(IOException::class)
        fun fromStr(`in`: DataInputStream): KeyBasedID {
            val representations = arrayOfNulls<ByteArray>(2)
            val hasKey = `in`.readBoolean()
            try {
                if (hasKey) {
                    val input = ByteArray(ID_LENGTH / 8 + 1)
                    `in`.readFully(input)
                    representations[1] = input
                    representations[0] = idFromInput(input, true)
                } else {
                    representations[0] = ByteArray(ID_LENGTH / 8)
                    `in`.readFully(representations[0])
                }
                return KeyBasedID(representations)
            } catch (e: Exception) {
                throw IOException(e)
            }
        }
    }

}
