package at.tugraz.iaik.p2p.kademlia.node

import at.tugraz.iaik.p2p.commons.CommonCrypto
import at.tugraz.iaik.p2p.commons.CryptoException
import java.io.DataInputStream
import java.io.DataOutputStream
import java.io.IOException
import java.util.*

open class SHA3ID
/**
 * Generate the NodeId from a given byte[]
 *
 * @param bytes
 */
@Throws(CryptoException::class) protected constructor(bytes: ByteArray, hash: Boolean) : Identifier(idFromInput(bytes,
                                                                                                                hash)) {
    @Throws(CryptoException::class) constructor(bytes: ByteArray) : this(bytes, true)

    override val type = Identifier.Type.SHA3

    override val numBits = ID_LENGTH


    override fun validate(bytes: ByteArray) {
        if (bytes.size != ID_LENGTH / 8) {
            throw IllegalArgumentException("Specified Data need to be " + ID_LENGTH / 8 + " characters long. " + "Data Given: '" + String(
                    bytes) + "'")

        }
    }

    override fun equals(other: Any?): Boolean {
        return if (other is SHA3ID) {
            Arrays.equals(this.keyBytes, other.keyBytes)
        } else false
    }

    override fun hashCode(): Int {
        var hash = 7
        hash = 83 * hash + Arrays.hashCode(this.keyBytes)
        return hash
    }

    /**
     * Checks the distance between this and another NodeId
     *
     * @param nid
     * @return The distance of this NodeId from the given NodeId
     */
    public override fun xor(nid: Identifier): SHA3ID {
        try {
            return SHA3ID(Identifier.xor(this.keyBytes, nid.keyBytes, 0), false)
        } catch (e: CryptoException) {
            throw RuntimeException(e)
        }

    }

    /**
     * Generates a NodeId that is some distance away from this NodeId
     *
     * @param distance in number of bits
     * @return NodeId The newly generated NodeId
     */
    override fun generateNodeIdByDistance(distance: Int): SHA3ID {
        return xor(SHA3ID(generateBytesByDitance(distance), false))
    }

    protected fun generateBytesByDitance(distance: Int): ByteArray {
        var dist = distance
        dist += 8
        val result = ByteArray(ID_LENGTH / 8)

        /* Since distance = ID_LENGTH - prefixLength, we need to fill that amount with 0's */
        val numByteZeroes = (ID_LENGTH - dist) / 8
        val numBitZeroes = 8 - dist % 8

        /* Filling byte zeroes */
        for (i in 0 until numByteZeroes) {
            result[i] = 0
        }

        /* Filling bit zeroes */
        val bits = BitSet(8)
        bits.set(0, 8)

        for (i in 0 until numBitZeroes) {
            /* Shift 1 zero into the start of the value */
            bits.clear(i)
        }
        bits.flip(0, 8)        // Flip the bits since they're in reverse order
        result[numByteZeroes] = bits.toByteArray()[0].toByte()

        /* Set the remaining keyBytes to Maximum value */
        for (i in numByteZeroes + 1 until result.size) {
            result[i] = java.lang.Byte.MAX_VALUE
        }
        return result
    }

    /**
     * Gets the distance from this NodeId to another NodeId
     *
     * @param to
     * @return Integer The distance
     */
    override fun getDistance(to: Identifier): Int {
        /**
         * Compute the xor of this and to
         * Get the index i of the first set bit of the xor returned NodeId
         * The distance between them is ID_LENGTH - i
         */
        return ID_LENGTH - xor(to).firstSetBitIndex(0)
    }

    @Throws(IOException::class)
    override fun toStream(out: DataOutputStream) {
        out.writeByte(type.id.toInt())
        out.write(this.keyBytes)
    }

    companion object {

        @Transient
        val ID_LENGTH = 32 * 8

        @Throws(CryptoException::class)
        private fun idFromInput(data: ByteArray, hash: Boolean): ByteArray {
            if (!hash) {
                return data
            }
            return CommonCrypto.sha3(data)
        }

        // @Override
        @Throws(IOException::class)
        fun fromStr(`in`: DataInputStream): SHA3ID {
            try {
                val representations = ByteArray(ID_LENGTH / 8)
                `in`.readFully(representations)
                return SHA3ID(representations, hash = false)
            } catch (e: Exception) {
                throw IOException(e)
            }
        }
    }

}
