@file:JvmName("BackRefs")
package at.tugraz.iaik.p2p.kademlia.routing

import at.tugraz.iaik.p2p.kademlia.node.Identifier
import at.tugraz.iaik.p2p.kademlia.node.KeyComparator
import at.tugraz.iaik.p2p.kademlia.node.Node
import at.tugraz.iaik.p2p.kademlia.node.SHA3ID
import java.math.BigInteger
import java.nio.ByteBuffer
import java.util.*
import kotlin.collections.set

data class BackRef(val virtualID: Identifier, val node: Node)

class BackRefComparator(private val key: BigInteger) : Comparator<BackRef> {
    override fun compare(o1: BackRef, o2: BackRef) = KeyComparator.compare(o1.virtualID.int, o2.virtualID.int, key)
}

class BackRefRoutingTable {
    private val virtualToNode = mutableMapOf<Identifier, Node>()
    private val nodeToVirtual = mutableMapOf<Node, MutableSet<Identifier>>()

    fun getNode(virtual: Identifier) = virtualToNode.get(virtual)
    fun getVirtual(node: Node) = nodeToVirtual.get(node)

    fun add(node: Node) {
        node.toBackRef().forEach { backref ->
            synchronized(nodeToVirtual) {
                virtualToNode[backref.virtualID] = backref.node
                if (!nodeToVirtual.containsKey(backref.node)) nodeToVirtual[backref.node] = mutableSetOf<Identifier>()
                nodeToVirtual[backref.node]!!.add(backref.virtualID)
            }
        }

    }



    fun remove(node: Node) {
        synchronized(nodeToVirtual) { nodeToVirtual.remove(node)?.let { it.forEach { virt -> virtualToNode.remove(virt) } } }
    }

    fun getClosestNodes(target: SHA3ID, numNodesRequired: Int): Set<Node> {
        val sortedSet = TreeSet(BackRefComparator(target.int))
        sortedSet.addAll(virtualToNode.map { BackRef(it.key, it.value) })

        val closest = mutableSetOf<Node>()

        /* Now we have the sorted set, lets get the top numRequired */
        for (n in sortedSet) {
            closest.add(n.node)
            if (closest.size == numNodesRequired) {
                break
            }
        }
        return closest
    }


    fun getSnapshot(): Map<Node, Set<Identifier>> = HashMap<Node, Set<Identifier>>(nodeToVirtual)
}

fun Node.toBackRef(): Set<BackRef> {
    val backrefs = mutableSetOf<BackRef>()
    backrefs.add(BackRef(nodeId,this))
    val buf = ByteBuffer.allocate(nodeId.numBits / 8 + 2)
    buf.put(nodeId.keyBytes)
    for (i in 0 until nonceRange) {
        buf.putShort(nodeId.numBits / 8, i.toShort())
        backrefs.add(BackRef(SHA3ID(buf.array()), this))
    }
    return backrefs
}